//=============================================================================
//  Add cautionary accidentals v0.1
//
//  Copyright (C)2012 Jörn Eichler (heuchi) 
//
//  This program is free software: you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation, either version 3 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program.  If not, see <http://www.gnu.org/licenses/>.
//=============================================================================

/* 
    features still missing for a plugin to fully work

 1. write access to Note.userAccidental

    currently Note.userAccidental can only be read. writing
    to it doesn't do anything

 2. Cursor.currentKeysig()

    the Cursor class should contain a member function to
    obtain the current key signature for the current staff
    at the current position of the cursor

 3. Note.writtenOnStaff

    using CTRL-SHIFT-UP/DOWN it is possible to move a note
    to a different staff.

    Something like Note.writtenOnStaff member would be needed, 
    that contains the number of the staff, on which the note actually
    appears to solve some special cases where the same note might
    appear several times in the same measure on different staves
*/

function init() {};

//=============================================================================
// function getLastStaffOfPart(startStaff)
//
// return the first staff that does not belong to the part to which 
// <startStaff> belongs.
//=============================================================================

function getLastStaffOfPart(startStaff) {
    var staffCounter=0;
    var partCounter=0;

    // print("gLSOP: called with "+startStaff);

    // skip staves until we reach our start

    while(partCounter<curScore.parts && staffCounter<startStaff) {
	var part=curScore.part(partCounter);
	staffCounter+=part.staves;
	// print("gLSOP: skipped "+part.staves+" staves for part "+partCounter);
	partCounter++;
    }

    if(startStaff==staffCounter) {
	// direct hit
	part=curScore.part(partCounter);
	//print("gLSOP: return value "+(startStaff+part.staves));
	return(startStaff+part.staves);
    } else if(startStaff<staffCounter) {
	// skipped the start, so we're already at the 
	// start of the next part
	//print("gLSOP: return value "+staffCounter);
	return(staffCounter);
    }

    // if we're here, there was a problem
    print("Error: gLSOP shouldn't get here!");
    return(-1);
}

//=============================================================================
// function parseStaves(startStaff,endStaff)
//
// do the actual work: find all notes in the selection and in the given
// staves that need an cautionary accidental
//
// this is needed for instruments (parts) that use more than one staff
// like pianos, because we also want extra accidentals if the same note
// in the same octave happens to be in a different staff of the same part
//
// PROBLEM: this cannot yet be fully implemented, since we don't know
//          in which staff a given note is printed. (see above)
//=============================================================================

function parseStaves(startStaff,endStaff) {
    //print("checkStaves called with "+startStaff+", "+endStaff);

    var cursor = new Cursor(curScore);
    cursor.goToSelectionStart();

    var selectionEnd = new Cursor(curScore);
    selectionEnd.goToSelectionEnd();

    /* we need to parse the score by measures
       so we keep a cursor, telling us where the
       next measure starts. We're inside the
       same measure as long as this holds:
       
       cursor.ticks()< nextMeasure.ticks()
    */

    nextMeasure = new Cursor(curScore);
    nextMeasure.goToSelectionStart();
    nextMeasure.nextMeasure();

    /* for each measure we create a table that contains 
       the actual 'noteName' of each 'noteClass'
      
       a 'noteClass' is the natural name of a space
       or line of the staff and the octave:
       C5, F6, B3 are 'noteClass'
    
       a 'noteName' would be C, F#, Bb for example
       (we don't need the octave here)

       curMeasureArray[<noteClass>] = <noteName>
    */

    var curMeasureArray= new Array();

    // we remember the same array of the previous measure
    var prevMeasureArray= new Array();

    // a color to mark the notes, that need an accidental
    // we cannot do it ourselves currently... :-(
    var red = new QColor(255,0,0);

    // parse the selection

    while(cursor.tick() < selectionEnd.tick()) {

	// check if still inside same measure
	if(!(cursor.tick() < nextMeasure.tick())) {
	    // measure change
	    //print("Next measure");
	    
	    var prevMeasureArray = curMeasureArray;
	    var curMeasureArray = new Array();
	    
	    nextMeasure.nextMeasure();
	}

	// check every staff and voice for notes:

	for(var staff=startStaff; staff<endStaff; ++staff) {
	    cursor.staff=staff;

	    for(var voice=0;voice<4;++voice) {
		cursor.voice=voice;

		if(cursor.isChord()) {
		    var chord=cursor.chord();
		    var n=chord.notes;

		    for(var i=0;i<n;i++) {
			var note = chord.note(i);
			
			// remember note for next measure
			
			var octave=Math.floor(note.pitch/12);
			var noteName = note.name;
			var noteClass = note.name.charAt(0)+octave;
		    
			/*print("found "+noteName
			      +" for ["+noteClass+"]"
			      +" with userAccidental="
			      +note.userAccidental
			      +" in Staff "+staff
			      +" voice "+voice);
			*/
			curMeasureArray[noteClass]=noteName;

			/* check if current note needs
			   cautionary

			   first we need to check if there
			   was a note of this noteClass
			   in the previous mesaure at all
			*/
			
			if(typeof prevMeasureArray[noteClass] 
			   !== 'undefined') {
			    // if so, did it contain a different noteName?
			    
			    if(prevMeasureArray[noteClass] != noteName) {
				/* this is a candidate
				   we need to check, if there is already
				   an accidental in front of the note
				   this can be done using the current
				   key signature.
				*/
				   
				// PROBLEM: we can't know the current keysig
				//          only Score.keysig is available
				//          as a work-around...
				
				var curKeysig=curScore.keysig;
				
				/* if the tpc of the note holds
				   
				   (12+keysig) < tpc < (20+keysig)
				   
				   the note doesn't get a necessary
				   accidental. So it needs a cautionary
				   if it has not already one (userAccidental)
				*/
				
				if((note.tpc > (curKeysig+12))
				   && (note.tpc < (curKeysig+20))) {
				    
				    // calculate needed accidental
				    
				    var neededAcc;
				    var showString;
				    
				    if(note.tpc < 6) {
					neededAcc=4; // double flat
					showString="bb";
				    } else if(note.tpc < 13) {
					neededAcc=2; // flat
					showString="b";
				    } else if(note.tpc < 20) {
					neededAcc=5; // natural
					showString="nat";
				    } else if(note.tpc < 27) {
					neededAcc=1; // sharp
					showString="#";
				    } else {
					neededAcc=3; // double sharp
					showString="##";
				    }
				    
				    if(note.userAccidental != neededAcc) {
					
					// needs a cautionary
					// for now mark it red
					note.color=new QColor(red);
					
					//print("This note needs cautionary");

					/*
					noteText = new Text(curScore);
					noteText.text = "Needs "+showString;
					noteText.yOffset=-2.0;
					
					cursor.putStaffText(noteText);
					*/
				    }
				}
			    }
			    // remove this entry in the old table
			    // since we don't want to reuse it
			    delete prevMeasureArray[noteClass];
			}
		    } // for on notes
		} // is chord?
	    } // for on voices
	} // for on staff

	// now go to next event
	cursor.next();
    }
    // done with selection
};

function run() 
{
    // check if we have a score
    if(typeof curScore === 'undefined')
	return;

    var cursor = new Cursor(curScore);
    cursor.goToSelectionStart();

    // check if there is a selection
    if(cursor.eos()) {
	return;
    }

    var selectionEnd = new Cursor(curScore);
    selectionEnd.goToSelectionEnd();

    var startStaff=cursor.staff;
    var endStaff=selectionEnd.staff;

    // go trough all staves of a part simultaneously
    // so find staves that belong to parts

    var curStartStaff=startStaff;

    while(curStartStaff<endStaff) {
	// find end staff for this part
	var curEndStaff=getLastStaffOfPart(curStartStaff);

	if(curEndStaff == -1) {
	    // something went wrong...
	    // better to do nothing than crashing...
	    return;
	}

	if(curEndStaff > endStaff) {
	    // user didn't select the whole part
	    // well, we hope they know what they're doing :-)
	    curEndStaff=endStaff;
	}

	// parse the selection

	//print("parsing "+curStartStaff+" to "+curEndStaff);

	parseStaves(curStartStaff,curEndStaff);

	// next iteration
	curStartStaff=curEndStaff;
    }
};

var mscorePlugin =
{
    majorVersion: 0,
    minorVersion: 1,
    menu: 'Plugins.Mark missing cautionary accidentals',
    init: init,
    run: run,
};

mscorePlugin;
